package ui

import (
	"strings"

	tea "github.com/charmbracelet/bubbletea"
)

const unknownCmdHelp = "Unknown command? Run /help to see available commands"

const cmdHelp = `/acn on                                      | Turn on the Tor ACN
/acn off                                     | Turn off the Tor ACN
/clear                                       | Clear the screen
/help                                        | Show this help
/profile create <name> <password> <password> | Create a new profile
/profile info                                | Show profile information
/profile unlock <password>                   | Unlock profile(s)
/start                                       | Show getting started guide
/quit                                        | Quit the program`

func isAcnOnCmd(cmds []string) bool {
	return cmds[0] == "acn" && cmds[1] == "on"
}

func isAcnOffCmd(cmds []string) bool {
	return cmds[0] == "acn" && cmds[1] == "off"
}

func isClearCmd(cmds []string) bool {
	return cmds[0] == "clear"
}

func handleCommand(cmd, hiddenInput string) tea.Msg {
	cmds := strings.Split(cmd, " ")

	switch {
	case cmds[0] == "help":
		return cmdMsg{output: strings.Split(cmdHelp, "\n")}

	case cmds[0] == "quit":
		return turnAcnOffMsg{quitProgram: true}

	case cmds[0] == "start":
		return showGettingStartedGuideMsg{}

	case cmds[0] == "profile":
		if cmds[1] != "unlock" && cmds[1] != "info" && cmds[1] != "create" {
			profileHelp := `Unknown "profile" command?
/profile create <name> <password> <password> | Create a new profile
/profile info                                | Show profile information
/profile unlock <password>                   | Unlock profile(s)`
			return cmdMsg{output: strings.Split(profileHelp, "\n")}
		}

		switch {
		case cmds[1] == "unlock":
			if len(cmds) != 3 {
				return cmdMsg{output: strings.Split(cmdHelp, "\n")}
			}

			return profileUnlockMsg{
				password: strings.TrimSpace(hiddenInput),
			}

		case cmds[1] == "info":
			return profileInfoMsg{}

		case cmds[1] == "create":
			if len(cmds) != 5 {
				return cmdMsg{output: strings.Split(cmdHelp, "\n")}
			}

			passwords := strings.Split(strings.TrimSpace(hiddenInput), " ")
			if passwords[0] != passwords[1] {
				return cmdMsg{output: []string{"Profile create: passwords do not match?"}}
			}

			return createProfileMsg{
				name:     cmds[2],
				password: passwords[1],
			}
		}

	case isAcnOnCmd(cmds):
		return turnAcnOnMsg{}

	case isAcnOffCmd(cmds):
		return turnAcnOffMsg{}

	case isClearCmd(cmds):
		return clearScreenMsg{}
	}

	return cmdMsg{output: []string{unknownCmdHelp}}
}

func hidePasswordInput(m *model) {
	val := strings.TrimSpace(m.input.Value())

	if len(val) == 0 {
		return
	}

	if string(val[0]) != "/" {
		return
	}

	cmds := strings.Split(val[1:], " ")
	if !(cmds[0] == "profile" && len(cmds) >= 2) {
		return
	}

	if cmds[1] == "create" && len(cmds) > 3 ||
		cmds[1] == "unlock" && len(cmds) > 2 {
		lastChar := string(val[len(val)-1])
		if lastChar != " " {
			m.hiddenInput += lastChar
			newCmd := []rune(val)
			newCmd[len(val)-1] = '*'
			m.input.SetValue(string(newCmd))
		}
	}
}
